#include <cstdlib>
#include <cassert>
#include <iostream>
#include "octree.hpp"

#define get16(x) ((x[1]<<8)|x[0])

typedef struct { 
  unsigned char       manufacturer;
  unsigned char       version;
  unsigned char       encoding;
  unsigned char       bitsperpixel;
  unsigned char       xmin[2],  ymin[2];
  unsigned char       xmax[2],  ymax[2];
  unsigned char       hDpi[2],  vDpi[2];
  unsigned char       colormap[48];
  unsigned char       reserved;
  unsigned char       nplanes;
  unsigned char       bitsPerLine[2];
  unsigned char       paletteInfo[2];
  unsigned char       hscreenSize[2], vscreenSize[2];
  unsigned char       filler[54];
}PCX_header;

int main(int argc, char *argv[])
{
    ////////////////////////////////////////////////////////////////////
    // Old C Pcx reader
    ////////////////////////////////////////////////////////////////////
    FILE* input;
    FILE* output;

    PCX_header     header;
    unsigned char* buffer;
    unsigned char* ptr;
    unsigned char* pptr;
    unsigned char* nextLine;
    unsigned int   w, h;

    unsigned int i, j, idx;
   
    unsigned char c, p;

    if(argc < 3) {
        fprintf(stderr,"Usage: %s input.pcx output.pcx\n",argv[0]);
        return 1;
    }

    /* Read pcx file */
    if(!(input = fopen(argv[1],"rb"))) {
        fprintf(stderr,"Error while loading %s\n",argv[1]);
        return 1;
    }

    fread(&header, 1, sizeof(PCX_header), input);

    if((header.bitsperpixel * header.nplanes) != 24) {
        fprintf(stderr,"%s must be a 24bpp pcx image\n",argv[1]);
        fclose(input);
        return 1;
    }

    w = get16(header.xmax) - get16(header.xmin) + 1;
    h = get16(header.ymax) - get16(header.ymin) + 1;

    buffer = (unsigned char*)malloc(w * h * 3 * sizeof(unsigned char));
    assert(buffer);

    ptr = buffer;
    for(i=0; i<h; ++i) {
		nextLine = ptr + w * 3;
		pptr = ptr;

		p = 0;
		do {
			fread(&c, 1, sizeof(unsigned char), input);
			if ((c & 0xC0) == 0xC0) {
				j = c & 0x3F;
				fread(&c, 1, sizeof(unsigned char), input);

				while (j--) {
					*pptr = c;
					pptr += 3;
					if (pptr >= nextLine) {
						p++;
						pptr = ptr + p;
					}
				}
			} else {
				*pptr = c;
				pptr += 3;
				if (pptr >= nextLine) {
					p++;
					pptr = ptr + p;
				}
			}
		} while (p < 3);

		ptr = nextLine;
    }

    fclose(input);
    ////////////////////////////////////////////////////////////////////

    printf("Reduce color\n");

    Quantizer::Octree tree(16);

    ptr = buffer;
    for(j=0; j<h; ++j) {
        for(i=0; i<w; ++i) {
            tree.Add(*ptr, *(ptr+1), *(ptr+2));
			ptr += 3;
        }
    }
    tree.Realize();
    
    unsigned int count;
    unsigned char *pal = tree.Palette(count);

    printf("Write pcx\n");

    ////////////////////////////////////////////////////////////////////
    // Write PCX
    ////////////////////////////////////////////////////////////////////
    header.manufacturer = 0x0a;
    header.version      = 5;
    header.encoding     = 1;
    header.bitsperpixel = 8;
    header.xmin[0] = header.xmin[1] = 0;
    header.ymin[0] = header.ymin[1] = 0;
    header.xmax[0] = (w-1) & 0xff;
    header.xmax[1] = ((w-1) >> 8) & 0xff;
    header.ymax[0] = (h-1) & 0xff;
    header.ymax[1] = ((h-1) >> 8) & 0xff;
    header.hDpi[0] = 0x2c;
    header.hDpi[1] = 0x01; 
    header.vDpi[0] = 0x2c;
    header.vDpi[1] = 0x01;
    memset(header.colormap, 0, 48);
    header.reserved  = 0;
    header.nplanes   = 1;
    header.bitsPerLine[0] = w & 0xff;
    header.bitsPerLine[1] = (w >> 8) & 0xff;
    header.paletteInfo[0] = 1;
    header.paletteInfo[1] = 0;
    header.hscreenSize[0] = w & 0xff;
    header.hscreenSize[1] = (w >> 8) & 0xff;
    header.vscreenSize[0] = h & 0xff;
    header.vscreenSize[1] = (h >> 8) & 0xff;
    memset(header.filler, 0, 54);

    /* Read pcx file */
    if(!(output = fopen(argv[2],"wb"))) {
        fprintf(stderr,"Error while loading %s\n",argv[2]);
        return 1;
    }

    fwrite(&header, 1, sizeof(PCX_header), output);

    ptr = buffer;
    for (j=0; j<h; j++) {
        for(i=0; i<w; ++i) {
            idx = tree.Index(*ptr, *(ptr+1), *(ptr+2));
            ptr += 3;

			if((idx & 0xc0) == 0xc0) {
				c = 0xc1;
				fwrite(&c, 1, 1, output);
			}
			c = idx;
			fwrite(&c, 1, 1, output);
		}
    }

    c = 0x0c;
    fwrite(&c, 1, 1, output);
    fwrite(pal, 3, count, output);
    if(count<256) {
		c = 0;
		for(i=count;i<256;++i) {
			fwrite(&c, 1, 1, output);
			fwrite(&c, 1, 1, output);
			fwrite(&c, 1, 1, output);
		}
    }
    fclose(output);

    delete [] pal;

    return 0;
}
