#ifndef __OCTREE_HPP__
#define __OCTREE_HPP__

#include <iostream>

namespace Quantizer {

    class Octree {
        public:
            // CTOR
            Octree(const unsigned int&, const unsigned int& = 8);
            // DTOR
            ~Octree();

            // Add a color to the octree
            void Add(const unsigned char&,
                     const unsigned char&, 
                     const unsigned char&);
            
            // Perform quantization
            void Realize();
            
            // Data accessors
            unsigned int ColorCount() const;
            unsigned int BitCount() const;
            // Compute palette
            unsigned char* Palette(unsigned int&);
            // Get palette index 
            unsigned int Index(const unsigned char&,
                               const unsigned char&,
                               const unsigned char&);

        private:
            // Octree node
            struct Node {
                // CTOR
                Node(const unsigned char&, Node*);
                // DTOR
                virtual ~Node();

				// Tells if the node has child
				virtual bool HasChildren() = 0;

				// Index
				unsigned char index;
				// Parent node
				Node* parent;
				// Previous node in the list
				Node* previous;
				// Next node in the list
				Node* next;
			};

            // Leaf node
            struct Leaf : public Node {
				// CTOR
				Leaf(const unsigned char&, Node*);
				// DTOR
				~Leaf();

				// Leaves don't have children
				bool HasChildren();

				// Add color to leaf	
				void Add(const unsigned char&,
						 const unsigned char&,
						 const unsigned char&);

				// Color
				unsigned long int color[3];
				// Reference count
				unsigned long int reference;
            };

			// Internal node
			struct Internal : public Node {
				// CTOR
				Internal(const unsigned char&, Node*);
				// DTOR
				~Internal();

				// Internal nodes at least have one child
				bool HasChildren();

				// Reduce
				Leaf* Reduce(Node**, unsigned int&);

				// Children
				Node* child[8];
			};

			// Add node to list
			void Add(const unsigned int&, Node*);

            // Tree root
            Node*  root;
            // Elligible nodes for reduction
            Node** candidates;
            // Number of colors
            unsigned int colorCount;
            // Current color count (leaf count)
			unsigned int leafCount;
            // Number of bits per color
            unsigned int bits;
    };        
          
};

#endif // __OCTREE_HPP__
